/*
 * Copyright 2017 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.diirt.datasource.internal;

import java.lang.reflect.Field;
import java.net.InetSocketAddress;
import java.security.AccessController;
import java.security.PrivilegedAction;

import com.aquenos.epics.jackie.client.ChannelAccessChannel;

/**
 * Utility that uses reflection to get the server address for a channel.
 * 
 * @author Sebastian Marsching
 */
public class ReflectionChannelServerAccess {

    /**
     * The only instance of this class.
     */
    public static final ReflectionChannelServerAccess INSTANCE = new ReflectionChannelServerAccess();

    private Field channelAccessChannelImplChannelField;
    private Field channelAccessPVServerConnectionServerAddressField;
    private Field channelLockField;
    private Field channelServerField;
    private Field serverConnectionField;
    private final boolean usable;

    /**
     * Returns the server address associated with a certain channel. This method
     * internally uses reflection to access private fields of the channel object
     * and other objects related to the channel. For this reason, this method
     * will fail with an {@link UnsupportedOperationException} if the server
     * address cannot be retrieved for some unknown reason. If the address
     * cannot be retrieved because the channel seems to be disconnected, an
     * {@link IllegalStateException} is thrown.
     * 
     * @param channel
     *            channel for which the server address should be retrieved.
     * @return server address associated with the channel.
     */
    public InetSocketAddress getServerAddress(ChannelAccessChannel channel) {
        if (!usable) {
            throw new UnsupportedOperationException(
                    "Reflective access failed.");
        }
        InetSocketAddress serverAddress = AccessController
                .doPrivileged((PrivilegedAction<InetSocketAddress>) () -> {
                    try {
                        return getServerAddressPrivileged(channel);
                    } catch (Throwable e) {
                        throw new UnsupportedOperationException(
                                "Reflective access failed", e);
                    }
                });
        if (serverAddress == null) {
            throw new IllegalStateException("The channel is not connected.");
        }
        return serverAddress;
    }

    // Private constructor so that only one instance can be created.
    private ReflectionChannelServerAccess() {
        this.usable = AccessController
                .doPrivileged((PrivilegedAction<Boolean>) () -> {
                    try {
                        initializePrivileged();
                        return Boolean.TRUE;
                    } catch (Throwable e) {
                        return Boolean.FALSE;
                    }
                });
    }

    private InetSocketAddress getServerAddressPrivileged(
            ChannelAccessChannel channel) throws IllegalAccessException {
        Object internalChannel = channelAccessChannelImplChannelField
                .get(channel);
        Object internalChannelLock = channelLockField.get(internalChannel);
        Object server;
        synchronized (internalChannelLock) {
            server = channelServerField.get(internalChannel);
        }
        // We use a return value of null in order to indicate that the server
        // object was null, which is usually caused by the channel not being
        // connected.
        if (server == null) {
            return null;
        }
        Object connection = serverConnectionField.get(server);
        Object serverAddress = channelAccessPVServerConnectionServerAddressField
                .get(connection);
        if (serverAddress == null) {
            throw new NullPointerException("Server address is null.");
        }
        return (InetSocketAddress) serverAddress;
    }

    private void initializePrivileged() throws ReflectiveOperationException {
        ClassLoader classLoader = ChannelAccessChannel.class.getClassLoader();
        Class<?> channelAccessChannelImplClass = Class.forName(
                "com.aquenos.epics.jackie.client.internal.ChannelAccessChannelImpl",
                true, classLoader);
        Field channelAccessChannelImplChannelField = channelAccessChannelImplClass
                .getDeclaredField("channel");
        channelAccessChannelImplChannelField.setAccessible(true);
        Class<?> channelClass = channelAccessChannelImplChannelField.getType();
        Field channelLockField = channelClass.getDeclaredField("lock");
        channelLockField.setAccessible(true);
        Field channelServerField = channelClass.getDeclaredField("server");
        channelServerField.setAccessible(true);
        Class<?> serverClass = channelServerField.getType();
        Field serverConnectionField = serverClass
                .getDeclaredField("connection");
        serverConnectionField.setAccessible(true);
        Class<?> channelAccessPVServerConnectionClass = serverConnectionField
                .getType();
        Field channelAccessPVServerConnectionServerAddressField = channelAccessPVServerConnectionClass
                .getDeclaredField("serverAddress");
        channelAccessPVServerConnectionServerAddressField.setAccessible(true);
        // If we made it here, we almost made it. Now we only have to check that
        // the serverAddress field is actually an InetSocketAddress.
        if (!InetSocketAddress.class.isAssignableFrom(
                channelAccessPVServerConnectionServerAddressField.getType())) {
            // If the type of the field does not match the expected type, we
            // cannot use the field.
            throw new ClassCastException(
                    "The field serverAddress exists, but it does not have the right type.");
        }
        // Now we have all objects that we need, so we can save them for later
        // use.
        this.channelAccessChannelImplChannelField = channelAccessChannelImplChannelField;
        this.channelLockField = channelLockField;
        this.channelServerField = channelServerField;
        this.serverConnectionField = serverConnectionField;
        this.channelAccessPVServerConnectionServerAddressField = channelAccessPVServerConnectionServerAddressField;
    }

}
